XB.network = {
    init: function() {
        this._resources = {};
        this._downloaders = {};
        this._logger = XB._base.application.core.Log4Moz.repository.getLogger(XB._base.loggersRoot + ".Network");
    },
    
    finalize: function() {
        for each (let resource in this._resources)
            try {
                resource.finalize();
            }
            catch (e) {
                this._logger.error("Error finalizing resource. " + XB._base.formatException(e));
            }
        
        for each (let downloader in this._downloaders)
            try {
                downloader.finalize();
            }
            catch (e) {
                this._logger.error("Error finalizing downloader. " + XB._base.formatException(e));
            }
        
        this._resources = null;
        this._downloaders = null;
    },
    
    getResource: function XBNet_getResource(requestData) {
        XB._base.runtime.ensureValueTypeIs(requestData, "RequestData");
        
        var resHash = requestData.hash;
        this._logger.debug("Looking for resource with hash \"" + resHash + "\"");
        var resource = this._resources[resHash];
        if (!resource) {
            this._logger.debug("Creating new resource");
            resource = new XB.network._Resource(requestData);
            this._resources[resHash] = resource;
        }
        return resource;
    },
    
    requestFinished: function XBNet_requestFinished(requestID) {
        return (requestID < this._requestID) && !(requestID in this._reqFinishSubs);
    },
    
    subscribeToRequestFinish: function XBNet_subscribeToRequestFinish(requestID, subscriber) {
        var subsArray = this._reqFinishSubs[requestID];
        if (!subsArray)
            this._reqFinishSubs[requestID] = [subscriber];
        else
            subsArray.push(subscriber);
    },
    
    ResourceDataType: {
        NRDT_CONTENT: 0,
        NRDT_HEADERS: 1,
        NRDT_STATUS: 2,
        NRDT_TRYID: 3
    },
    
    _consts: {
    },
    _reqFinishSubs: {},
    _resources: null,
    _downloaders: null,
    _logger: null,
    _requestID: 1,
    
    _disposeResource: function XBNet_disposeResource(netResource) {
        var resHash = netResource.requestDataHash;
        netResource.finalize();
        delete this._resources[resHash];
    },
    
    _getDownloader: function XBNet_getDownloader(requestData) {
        XB._base.runtime.ensureValueTypeIs(requestData, "RequestData");
        
        var dlHash = this._dlHash(requestData);
        var downloader = this._downloaders[dlHash];
        if (!downloader) {
            downloader = new XB.network._ResourceDownloader(requestData);
            this._downloaders[dlHash] = downloader;
        }
        return downloader;
    },
    
    _getNewRequestID: function XBNet_getNewRequestID() {
        var requestID = this._requestID++;
        this._reqFinishSubs[requestID] = undefined;
        return requestID;
    },
    
    _dlHash: function XBNet_dlHash(requestData) {
        return [requestData.url, requestData.method].join("#");
    },
    
    _onDownloadFinished: function XBNet_onDownloadFinished(requestID) {
        var subscribers = this._reqFinishSubs[requestID];
        if (subscribers) {
            for each (let subscriber in subscribers)
                try {
                    subscriber.onRequestFinished(requestID);
                }
                catch (e) {
                    XB._base.logger.error(XB._base.formatException(e));
                }
        }
        
        delete this._reqFinishSubs[requestID];
    }
};

XB.network._Resource = function XBNetResource_constructor(requestData) {
    this._timer = XB._Cc["@mozilla.org/timer;1"].createInstance(XB._Ci.nsITimer);
    this._subscribtions = {};
    this._downloader = XB.network._getDownloader(requestData);
    this._requestData = requestData;
    this._storageKey = this._makeStorageKey(requestData);
    this._data = {};
    if (this._loadFromCache()) {
        if (this._currDataTime !== undefined) {
            var dataExpirationInterval = (this._currDataTime + this._requestData.expirationInterval) * 1000 -
                                         (new Date()).getTime();
            dataExpirationInterval = Math.max(dataExpirationInterval, 0);
            this._timer.initWithCallback(this, dataExpirationInterval, this._timer.TYPE_REPEATING_SLACK);
            
            var extraUpdateInterval = (this._currDataTime + this._requestData.updateInterval) * 1000 -
                                     (new Date()).getTime();
            extraUpdateInterval = Math.max(extraUpdateInterval, 0);
            XB.network._logger.debug("Resource set extra update interval to " + extraUpdateInterval);
            this._extraUpdTimer = XB._Cc["@mozilla.org/timer;1"].createInstance(XB._Ci.nsITimer);
            this._extraUpdTimer.initWithCallback(this, extraUpdateInterval, this._timer.TYPE_ONE_SHOT);
        }
    }
};

XB.network._Resource.prototype = {
    constructor: XB.network._Resource,
    
    requestData: function XBNetResource_requestData(subscriber, dataType) {
        this._checkSubscription(subscriber, dataType);
        return this._getDataPart(dataType);
    },
    
    update: function XBNetResource_update(bypassCache) {
        return this._downloader.update(!!bypassCache);
    },
    
    unsubscribe: function XBnetResource_unsubscribe(subscriber, dataType) {
        if (!subscriber || (typeof subscriber.getEffectiveID !== "function") || (dataType === undefined) )
            throw new TypeError(this._consts.ERR_WRONG_SUBSCRIBER_INTF);
        var subHash = subscriber.getEffectiveID() + "#" + dataType;
        delete this._subscribtions[subHash];
        if (sysutils.isEmptyObject(this._subscribtions))
            XB.network._disposeResource(this);
    },
    
    finalize: function XBNetResource_finalize() {
        if (this._timer)
            this._timer.cancel();
        this._timer = null;
        if (this._extraUpdTimer)
            this._extraUpdTimer.cancel();
        this._extraUpdTimer = null;
        this._downloader.unsubscribe(this);
        this._downloader = null;
        this._requestData = null;
    },
    
    onDownloaderFinished: function XBNetResource_onDownloaderFinished() {
        if (this._updateFromDownloader()) {
            try {
                this._resetTimer();
                var cacheData = {
                        status: this._downloader.data.status,
                        headers: this._downloader.data.headers,
                        body: this._downloader.data.bodyText
                    };
                XB.cache.store(this._storageKey, cacheData);
            }
            finally {
                this._notifySubscribers();
            }
        }
    },
    
    get requestDataHash() {
        return this._requestData.hash;
    },
    
    _consts: {
        ERR_WRONG_SUBSCRIBER_INTF: "Wrong subscriber interface"
    },
    _timer: null,
    _extraUpdTimer: null,
    _subscribtions: null,
    _requestData: null,
    _storageKey: undefined,
    _downloader: null,
    _data: null,
    _subscribed: false,
    _currDataTime: undefined,
    
    _loadFromCache: function XBNetResource_loadFromCache() {
        try {
            var storeTime = {};
            var responseData = XB.cache.findData(this._storageKey, this._requestData.expirationInterval, storeTime);
            if (responseData) {
                var responseData = XB._base.application.core.JSON.parse(responseData);
                var content = this._isXML? this._tryConvertToXML(responseData.body): responseData.body;
                if (this._validate(responseData.status, content, responseData.headers)) {
                    this._data.headers = responseData.headers;
                    this._data.content = content;
                    this._data.status = responseData.status;
                    
                    this._currDataTime = storeTime.timestamp;
                    return true;
                }
            }
        }
        catch (e) {
            XB.network._logger.error("Couldn't load cached response. " + XB._base.formatException(e));
        }
        
        return false;
    },
    
    _getDataPart: function XBNetResource_getDataPart(dataType) {
        switch (dataType) {
            case XB.network.ResourceDataType.NRDT_CONTENT:
                return this._data.content;
            case XB.network.ResourceDataType.NRDT_HEADERS:
                return this._data.headers;
            case XB.network.ResourceDataType.NRDT_STATUS:
                return this._data.status;
            case XB.network.ResourceDataType.NRDT_TRYID:
                return NaN;
        }
        return undefined;
    },
    
    _checkSubscription: function XBNetResource_checkSubscription(subscriber, dataType) {
        this._subscribe(subscriber, dataType);
        if (!this._subscribed) {
            if (this._downloader.data.status != -1)
                this._updateFromDownloader();
            this._downloader.subscribe(this, this._requestData.updateInterval);
            if (sysutils.isEmptyObject(this._data))
                this._downloader.update();
            this._subscribed = true;
        }
    },
    
    _subscribe: function XBNetResource_subscribe(subscriber, dataType) {
        if (typeof subscriber.update !== "function")
            throw new TypeError(this._consts.ERR_WRONG_SUBSCRIBER_INTF);
        var subHash = subscriber.getEffectiveID() + "#" + dataType;
        this._subscribtions[subHash] = {subscriber: subscriber, dataType: dataType};
    },
    
    _updateFromDownloader: function XBNetResource_updateFromDownloader() {
        XB.network._logger.debug("_updateFromDownloader");
        var data = this._downloader.data;
        var content = this._isXML? this._tryConvertToXML(data.bodyText): data.bodyText;
        if (this._validate(data.status, content, data.headers)) {
            this._data.content = content;
            this._data.headers = data.headers;
            this._data.status = data.status;
            
            this._currDataTime = parseInt((new Date()).getTime() / 1000);
            return true;
        }
        return false;
    },
    
    _validate: function XBNetResource_validate(status, content, headers) {
        XB.network._logger.debug("Response status:" + status + ". Valid: " +
                                 this._requestData.statusRange.start + ".." + this._requestData.statusRange.end);
        var statusIsValid = (this._requestData.statusRange.start <= status && status <= this._requestData.statusRange.end);
        XB.network._logger.debug("Status is valid: " + statusIsValid);
        var structIsValid = true;
        if (this._isXML)
            try {
                if (!content)
                    throw new Error("Content is not XML");
                XB.network._logger.trace("Content xml root: " + content.documentElement.nodeName);
                var queryResult = XB._base.runtime.queryXMLDoc(this._requestData.xpathExpression, content);
                if ( ((queryResult instanceof Array) && !queryResult.length) || !queryResult )
                    throw new Error("Empty validation query result");
            }
            catch (e) {
                XB.network._logger.debug("Content validation failed. " + XB._base.formatException(e));
                structIsValid = false;
            }
        
        XB.network._logger.debug("Structure is valid: " + structIsValid);
        return (statusIsValid && structIsValid);
    },
    
    _resetTimer: function XBNetResource_resetTimer() {
        this._timer.cancel();
        this._timer.initWithCallback(this, this._requestData.expirationInterval * 1000, this._timer.TYPE_REPEATING_SLACK);
    },
    
    _notifySubscribers: function XBNetResource_notifySubscribers() {
        for each (let subscription in this._subscribtions) {
            try {
                subscription.subscriber.update(this, this._getDataPart(subscription.dataType));
            }
            catch (e) {
                XB.network._logger.error(XB._base.formatException(e));
            }
        }
    },
    
    _tryConvertToXML: function XBNetResource_tryConvertToXML(text) {
        if (!text)
            return null;
        
        var xmlDoc = XB._base.domParser.parseFromString(text, "text/xml");
        
        if (xmlDoc.documentElement.localName == "parsererror") {
            XB.network._logger.debug("Text is not XML");
            
            try {
                var jsObj = XB._base.application.core.JSON.parse(text);
                throw new Error("JSON to XML conversion is not implemented");
                
                return xmlDoc;
            }
            catch (e) {
                XB.network._logger.debug("Failed converting from JSON to XML. " + XB._base.formatException(e));
            }
            
            try {
                throw new Error("CSV to XML conversion is not implemented");
                
                return xmlDoc;
            }
            catch (e) {
                XB.network._logger.debug("Failed converting from CSV to XML. " + XB._base.formatException(e));
            }
            
            return null;
        }
        
        return xmlDoc;
    },
    
    _makeStorageKey: function XBNetResource_makeStorageKey(requestData) {
        return [requestData.url, requestData.method, requestData.format,
                requestData.statusRange.start, requestData.statusRange.end, requestData.xpathExpression].join("#")
    },
    
    get _isXML() {
        return this._requestData.format !== XB.types.RequestData.Format.FMT_TEXT;
    },
    
    notify: function XBNetResource_notify(timer) {
        if (timer === this._extraUpdTimer) {
            this.update(true);
        }
        else {
            this._timer.cancel();
            this._data = {};
            this._notifySubscribers();
        }
    }
};

XB.network._ResourceDownloader = function(requestData) {
    XB._base.runtime.ensureValueTypeIs(requestData, "RequestData");
    this._requestData = requestData;
    this._data = {
        status: -1,
        bodyText: "",
        headers: null};
    this._subscribtions = [];
    this._timer = XB._Cc["@mozilla.org/timer;1"].createInstance(XB._Ci.nsITimer);
    this._logger = XB._base.application.core.Log4Moz.repository.getLogger(XB._base.loggersRoot + ".HTTPDL" + (new Date()).getTime());
}

XB.network._ResourceDownloader.prototype = {
    constructor: XB.network._ResourceDownloader,
    
    update: function XBResDownloader_update(bypassCache) {
        this._sendRequest(!!bypassCache);
        this._timer.cancel();
        return this._currRequestID;
    },
    
    subscribe: function XBResDownloader_subscribe(subscriber, interval) {
        if ( !(subscriber instanceof XB.network._Resource) )
            throw new TypeError(this._consts.ERR_WRONG_SUBSCRIBER_INTF);
        if (interval < this._consts.MIN_INTERVAL)
            throw new Error(this._consts.ERR_SMALL_INTERVAL);
        
        var subscribtion = null;
        for (var i = this._subscribtions.length; i--; )
            if (this._subscribtions[i].subscriber == subscriber) {
                subscribtion = this._subscribtions[i];
                break;
            }
        
        if (!subscribtion) {
            subscribtion = {
                subscriber: subscriber,
                interval: interval
            };
            
            this._subscribtions.push(subscribtion);
        } else {
            subscribtion.interval = interval;
        }
        
        if (interval < this._currentInterval)
            this._currentInterval = interval;
        
        if (this._subscribtions.length == 1)
            this._resetTimer(this._currentInterval);
        
        this._logger.debug("Added subscriber. New interval is " + this._timer.delay / 1000 + "s");
    },
    
    unsubscribe: function XBResDownloader_unsubscribe(subscriber) {
        for (var i = this._subscribtions.length; i--; )
            if (this._subscribtions[i].subscriber === subscriber) {
                this._subscribtions.splice(i, 1);
                this._updateInterval();
                this._logger.debug("Removed subscriber. New interval is " + this._timer.delay / 1000 +
                                   "s. Subscriptions: " + this._subscribtions.length);
                break;
            }
    },
    
    get numSubscribers() {
        return this._subscribtions.length;
    },
    
    get data() {
        return this._data;
    },
    
    finalize: function XBResDownloader_finalize() {
        this._timer.cancel();
        this._requestData = null;
        this._subscribtions = null;
    },
    
    _consts: {
        MIN_INTERVAL: 5,
        ERR_SMALL_INTERVAL: "The requested resource update interval is too small",
        ERR_WRONG_SUBSCRIBER_INTF: "Subscriber does not support the required interface"
    },
    _currRequestID: undefined,
    _requestData: null,
    _data: null,
    _subscribtions: null,
    _timer: null,
    _logger: null,
    
    _currentInterval: Number.POSITIVE_INFINITY,
    
    _resetTimer: function XBResDownloader_resetTimer(newInterval) {
        this._timer.cancel();
        if (newInterval != undefined && newInterval != Number.POSITIVE_INFINITY) {
            this._timer.initWithCallback(this, newInterval * 1000, this._timer.TYPE_ONE_SHOT);
        }
        else {
            try {
                this._timer.delay = 0;
            } catch (e) {}
        }
    },
    
    _updateInterval: function XBResDownloader_updateInterval() {
        var interval = Number.POSITIVE_INFINITY;
        for (var i = this._subscribtions.length; i--; )
            if (this._subscribtions[i].interval < interval)
                interval = this._subscribtions[i].interval;
        
        this._resetTimer(interval);
        this._currentInterval = interval;
    },
    
    _sendRequest: function XBResDownloader_sendRequest(bypassCache) {
        this._logger.debug("_sendRequest " + [this._requestData.method, this._requestData.url, bypassCache]);
        try {
            this._currRequestID = XB.network._getNewRequestID();
            var request = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
            request.mozBackgroundRequest = true;
            request.open(this._requestData.method, this._requestData.url, true);
            if (bypassCache) {
                request.setRequestHeader("If-Modified-Since", "Sat, 1 Jan 2000 00:00:00 GMT");
            }
            
            var target = request.QueryInterface(Ci.nsIDOMEventTarget);
            target.addEventListener("load", this, false);
            target.addEventListener("error", this, false);
            request.send(null);
        }
        catch (e) {
            this._currRequestID = undefined;
            this._resetTimer(this._currentInterval);
            this._logger.error("Failed sending request. " + XB._base.formatException(e));
            this._logger.trace("Resetting timer.");
        }
    },
    
    _handleResponse: function XBResDownloader_handleResponse(request) {
        try {
            this._data.status = request.status;
            this._data.bodyText = request.responseText;
            this._data.headers = {};
            
            for (var i = this._subscribtions.length; i--; )
                try {
                    this._subscribtions[i].subscriber.onDownloaderFinished();
                }
                catch (e) {
                    this._logger.error("Failed notifying downloader subscriber. " + XB._base.formatException(e));
                    if (e.stack)
                        this._logger.debug(e.stack);
                }
        }
        catch(e) {
            this._logger.error("Error in _ResourceDownloader::_handleResponse. " + XB._base.formatException(e));
        }
        finally {
            var RID = this._currRequestID;
            this._currRequestID = undefined;
            this._resetTimer(this._currentInterval);
            XB.network._onDownloadFinished(RID);
        }
    },
    
    notify: function XBResDownloader_notify() {
        this._sendRequest();
    },
    
    handleEvent: function XBResDownloader_handleEvent(event) {
        switch(event.type) {
            case "load":
            case "error":
                this._handleResponse(event.target);
                break;
        }
    }
}
