XB.WidgetPackage = function XBWPackage_constructor(rootDir) {
    if ( !(rootDir instanceof XB._Ci.nsIFile) )
        throw new TypeError(XB._base.consts.ERR_FILE_REQUIRED);
    if ( !rootDir.isDirectory() )
        throw new TypeError(this._consts.ERR_NOT_A_DIR);
    this._rootDir = rootDir;
    this._name = this._rootDir.leafName;
    this._logger = XB._base.application.core.Log4Moz.repository.getLogger(XB._base.loggersRoot + ".Package_" + this._name);
    this._bestLocaleName = this._getBestLocaleName();
    
    this._xbProtocolHandler = Cc[PROTOCOL_XB_CONTRACTID].getService(Ci.nsIProtocolHandler).wrappedJSObject;
    this._xbProtocolHandler.addDataProvider(this);
    
    this._libs = {};
    this._loadLibFiles();
};

XB.WidgetPackage.prototype = {
    constructor: XB.WidgetPackage,
    
    get name() {
        return this._name;
    },
    
    getLibrary: function XBWidgetPackage_getLibrary(name) {
        var library = this._libs[name];
        if (!library)
            throw new Error(this._consts.ERR_NO_LIB);
        return library;
    },
    
    getFile: function XBWidgetPackage_getFile(path) {
        if (!sysutils.isString(path))
            throw new TypeError(this._consts.ERR_STRING_EXPECTED);
        var file = this._rootDir.clone();
        var pathComponents = path.split("/");
        for each(let pathComponent in pathComponents) {
            file.append(pathComponent);
            if (file.isSymlink() || file.contains(this._rootDir, false))
                throw new Error(this._consts.ERR_ACCESS_DENIED);
        }
        file.normalize();
        return file;
    },
    
    getXBURL: function XBPkg_getXBURL(relPath) {
        var spec = "xb://content/" + relPath + ";" + this.providerId;
        return spec;
    },
    
    finalize: function () {
        for (let libName in this._libs)
            try {
                this._libs[libName].clear();
            }
            catch (e) {
                this._logger.warn("Couldn't clear loaded library " + libName + ". " + XB._base.formatException(e));
            }
        this._libs = {};
        
        this._xbProtocolHandler.removeDataProvider(this);
    },
    
    get providerId() {
        return this.name + "." + XB._base.application.name;
    },
    
    newChannel: function(aURI) {
        var path = this._resolveURI(aURI);
        var file = this.getFile(path);
        
        var filesStream = XB._Cc["@mozilla.org/network/file-input-stream;1"]
            .createInstance(XB._Ci.nsIFileInputStream);
        filesStream.init(file, -1, 0, 0);
        
        var channel = XB._Cc["@mozilla.org/network/input-stream-channel;1"]
            .createInstance(XB._Ci.nsIInputStreamChannel)
                .QueryInterface(XB._Ci.nsIChannel);
        channel.setURI(aURI);
        channel.originalURI = aURI;
        channel.contentStream = filesStream;
        channel.contentCharset = "UTF-8";
        
        return channel;
    },
    
    _consts: {
        PKG_PATH_LOCALE: "locale",
        PKG_PATH_CONTENT: "content",
        PKG_PATH_SKIN: "skin",

        ERR_NOT_A_DIR: "Root must be a directory",
        ERR_NO_LIB: "No such library",
        ERR_ACCESS_DENIED: "Attempt to access a file outside the package directory",
        ERR_STRING_EXPECTED: "String expected"
    },
    _rootDir: null,
    _libs: null,
    _name: undefined,
    _bestLocaleName: undefined,
    _logger: null,
    
    _loadLibFiles: function XBWidgetPackage_loadLibFiles() {
        var entries = this._rootDir.directoryEntries;
        while (entries.hasMoreElements()) {
            var file = entries.getNext().QueryInterface(XB._Ci.nsIFile);
            if (file.isFile()) {
                if ( !file.leafName.match(/\.xml$/ig) || !file.isReadable() ) continue;
                
                var lib = new XB.WidgetLibrary(this);
                try {
                    var xmlDoc = this._loadXBDoc(this.getXBURL(file.leafName));
                    lib.loadLibDoc(xmlDoc);
                    var libName = file.leafName.substr(0, file.leafName.length - 4);
                    this._libs[libName] = lib;
                }
                catch (e) {
                    this._logger.warn("Couldn't load library file " + file.leafName + ". " + XB._base.formatException(e));
                }
            }
        }
    },
    
    _loadXBDoc: function XBWidgetPackage_loadXBDoc(uri) {
        var ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        var request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance();
        request.open('GET', uri, false);
        request.send(null);
        return request.responseXML;
    },
    
    _resolveURI: function XBWidgetPackage_resolveURI(URI) {
        var path = URI.path,
            host = URI.host;
        
        if (host) {
            switch (host) {
                case this._consts.PKG_PATH_LOCALE:
                    path = "/" + host + "/" + this._bestLocaleName + path;
                    this._logger.debug("Resulting path is: " + path);
                    break;
                case this._consts.PKG_PATH_SKIN:
                case this._consts.PKG_PATH_CONTENT:
                    break;
                default:
                    path = '/' + host + path;
            }
        }
        
        return path;
    },
    
    _getBestLocaleName: function XBWidgetPackage_getBestLocaleName() {
        function parseLocale(locale) {
            var components = locale.match(/([a-z]{2})(-([A-Z]{2})(-(\w{2,5}))?)?/);
            if (components)
                return {
                    language: components[1],
                    country: components[3],
                    region: components[5]
                };
            return null;
        }
        
        const localeComponentsWeights = {
            language: 4,
            country: 2,
            region: 1
        };
        
        var pretenderLocaleName = "en-US",
            maxWeight = 1;
        
        var appLocaleString = XB._base.application.localeString;
        var appLocale = parseLocale(appLocaleString);
        
        var localeDir = this._rootDir.clone();
        localeDir.append(this._consts.PKG_PATH_LOCALE);
        var entries = localeDir.directoryEntries;
        while (entries.hasMoreElements()) {
            var file = entries.getNext().QueryInterface(XB._Ci.nsIFile);
            if (file.isDirectory()) {
                var dirName = file.leafName;
                if (dirName == appLocaleString) {
                    pretenderLocaleName = dirName;
                    break;
                }
                
                var candidateLocale = parseLocale(dirName);
                if (!candidateLocale)
                    continue;
                
                var candidateLocaleWeight = 0;
                for (let component in localeComponentsWeights) {
                    var candidateComponent = candidateLocale[component];
                    if (candidateComponent === undefined) continue;
                    this._logger.trace("component " + component);
                    if (candidateComponent === appLocale[component]) {
                        candidateLocaleWeight += localeComponentsWeights[component];
                    }
                }
                
                if (candidateLocaleWeight > maxWeight) {
                    maxWeight = candidateLocaleWeight;
                    pretenderLocaleName = dirName;
                }
            }
        }
        this._logger.info("pretenderLocaleName " + pretenderLocaleName);
        return pretenderLocaleName;
    }
};
