XB.WidgetLibrary = function XBWidgetLib_constructor(pkg) {
    if ( !(pkg instanceof XB.WidgetPackage) )
        throw new TypeError(this._consts.ERR_WPACKAGE_REQUIRED);
    this._package = pkg;
    this._logger = XB._base.application.core.Log4Moz.repository.getLogger(XB._base.loggersRoot + ".WL");
    this._widgetsMap = {};
};

XB.WidgetLibrary.prototype = {
    constructor: XB.WidgetLibrary,
    
    get availableWidgetsIDs() {
        var result = [];
        for (let protoID in this._widgetsMap)
            result.push(protoID);
        return result;
    },
    
    getWidgetProto: function XBWidgetLib_getProto(protoSrcID) {
        this._logger.debug("Looking for widget prototype " + protoSrcID);
        var proto = this._widgetsMap[protoSrcID];
        if (!proto)
            throw new Error(this._consts.ERR_UNKNOWN_WIDGET + ( " (" + protoSrcID + ")" ));
        return proto;
    },
    
    clear: function() {
    },
    
    loadLibFile: function(libFile) {
        this._logger.debug("Loading XB library from " + libFile.path);
        this.loadLibDoc( sysutils.readXMLDocument(libFile) );
    },
    
    loadLibDoc: function(libDoc) {
        if (libDoc.documentElement.localName != this._consts.STR_LIB_NODE_NAME)
            throw new TypeError(XB._base.consts.ERR_INVALID_DOC_STRUC);
        
        let childNodes = libDoc.documentElement.childNodes;
        for (let subIndex = 0, len = childNodes.length; subIndex < len; subIndex++) {
            var subNode = childNodes[subIndex];
            if (subNode.nodeType != subNode.ELEMENT_NODE) continue;
            
            switch (subNode.localName) {
                case XB._base.consts.STR_WIDGET_ELEMENT_NAME:
                    try {
                        this._parseWidgetElement(subNode);
                    }
                    catch (e) {
                        this._logger.error(this._consts.ERR_PARSING_WIDGET + ". " + XB._base.formatException(e));
                    }
                    break;
                case "data":
                    break;
                default:
                    this._logger.warn(this._consts.WARN_UNKNOWN_TOKEN + " \"" + subNode.localName + "\"");
                    break;
            }
        }
    },
    
    get libPackage() {
        return this._package;
    },
    
    _consts: {
        STR_LIB_NODE_NAME: "library",
        STR_WIDGET_ID_ATTR_NAME: "id",
        STR_WIDGET_NAME_ATTR_NAME: "name",
        STR_FUNC_PARAM_ELEM_NAME: "param",
        
        WARN_UNKNOWN_TOKEN: "Unknown library token",
        WARN_WIDGET_ALREADY_REGISTERED: "Widget with this ID is already registered",
        WARN_FUNC_IGNORED: "Function was ignored because of inappropriate syntax",
        
        ERR_NO_WIDGET_ID: "No widget ID",
        ERR_NO_WIDGET_NAME: "No widget name",
        ERR_NO_DATA_NAME: "Unnamed data declaration",
        ERR_NO_VAR_NAME: "Unnamed variable declaration",
        ERR_NO_SETTING_NAME: "Unnamed setting",
        ERR_PARSING_WIDGET: "An error occured while parsing widget",
        ERR_PARSING_FUNC_NODE: "An error occured while parsing node",
        ERR_UNKNOWN_WIDGET: "Widget with this ID is not registered",
        ERR_INVALID_SCOPE: "Invalid scope definition",
        ERR_WPACKAGE_REQUIRED: "Widget package required",
        ERR_UNKNOWN_VALUE_TYPE: "Unknown value type"
    },
    _logger: XB._noopLogger,
    _widgetsMap: null,
    _package: null,
    _nodeUID: 1,
    
    _parseWidgetElement: function XBWidgetLib_parseWidgetElement(widgetElement) {
        var widgetID = widgetElement.getAttribute(this._consts.STR_WIDGET_ID_ATTR_NAME);
        if (!widgetID)
            throw new Error(this._consts.ERR_NO_WIDGET_ID);
        var widgetName = widgetElement.getAttribute(this._consts.STR_WIDGET_NAME_ATTR_NAME);
        if (!widgetName)
            throw new Error(this._consts.ERR_NO_WIDGET_NAME);
        
        if (this._widgetsMap[widgetID]) {
            this._logger.warn(this._consts.WARN_WIDGET_ALREADY_REGISTERED + " (" + widgetID + ")");
            return;
        }
        
        var widgetProto = new XB.WidgetLibrary.WidgetPrototype(widgetElement, this);
        
        var treeWalker = widgetElement.ownerDocument
            .createTreeWalker(widgetElement, XB._Ci.nsIDOMNodeFilter.SHOW_ALL, this._nodeFilter, true);
        
        if (treeWalker.firstChild()) {
            do {
                var subNode = treeWalker.currentNode;
                if (subNode.namespaceURI == XB._base.consts.STR_FUNCTIONAL_NS) {
                    switch (subNode.localName) {
                        case "data":
                            this._parseWidgetData(widgetProto, treeWalker);
                            break;
                        case "variable":
                            this._parseWidgetVar(widgetProto, treeWalker);
                            break;
                        case "setting":
                            this._parseWidgetSetting(widgetProto, treeWalker);
                            break;
                        default:
                            this._logger.warn(this._consts.WARN_FUNC_IGNORED + " (" + subNode.localName + ")");
                    }
                }
                else
                    this._parseElement(treeWalker, null, widgetProto);
            }
            while (treeWalker.nextSibling());
        }
        this._widgetsMap[widgetID] = widgetProto;
        this._logger.debug("Parsed widget prototype " + widgetProto.getID());
    },
    
    _parseWidgetData: function XBWidgetLib_parseWidgetData(parentWidget, treeWalker) {
        var dataNode = treeWalker.currentNode;
        var dataName = dataNode.getAttribute("name");
        if (!dataName)
            throw new XB.WidgetLibrary.EParseError(this._consts.ERR_NO_DATA_NAME, dataNode.nodeName);
        this._logger.trace("Found widget scope data " + dataName);
        
        var nodeID = this._genNodeUID();
        var dataNode = new XB._base.FuncNodeProto(nodeID, XB.functions.CN_concat);
        this._parseElement(treeWalker, dataNode, parentWidget);
        
        parentWidget.registerData(dataName, dataNode);
    },
    
    _evalScope: function XBWidgetLib_evalScope(scopeName) {
        switch (scopeName) {
            case "widget":
                return XB._base.consts.ENUM_SCOPE_WIDGET;
            case "instance":
                return XB._base.consts.ENUM_SCOPE_INSTANCE;
            default:
                throw new Error(this._consts.ERR_INVALID_SCOPE);
        }
    },
    
    _parseWidgetVar: function XBWidgetLib_parseWidgetVar(parentWidget, treeWalker) {
        var varNode = treeWalker.currentNode;
        var varName = varNode.getAttribute("name");
        if (!varName)
            throw new XB.WidgetLibrary.EParseError(this._consts.ERR_NO_VAR_NAME, varNode.nodeName);
        this._logger.trace("Found variable " + varName);
        var initialValue = XB.types.empty;
        var varScope = this._evalScope(varNode.getAttribute("scope"));
        var persist = (varNode.getAttribute("persist") == "true");
        parentWidget.registerVariable(varName, varScope, persist, initialValue);
    },
    
    _parseWidgetSetting: function XBWidgetLib_parseWidgetSetting(parentWidget, treeWalker) {
        var settingNode = treeWalker.currentNode;
        var settingName = settingNode.getAttribute("name");
        if (!settingName)
            throw new XB.WidgetLibrary.EParseError(this._consts.ERR_NO_SETTING_NAME, settingNode.nodeName);
        
        var settingScope = this._evalScope(settingNode.getAttribute("scope"));
        var defaultValue = settingNode.getAttribute("default");
        this._logger.trace("Found widget setting " + settingName + ", scope " + settingScope + ", default " + defaultValue);
        parentWidget.registerSetting(settingName, settingScope, defaultValue);
    },
    
    _parseElement: function XBWidgetLib_parseElement(treeWalker, parentCalcNode, parentWidget) {
        this._logger.trace("Checking node " + treeWalker.currentNode.nodeName + " children");
        var calcNodeArgs = [];
        
        if (treeWalker.firstChild())
            try {
                do {
                    var widgetSubNode = treeWalker.currentNode;
                    
                    if (parentCalcNode == null && widgetSubNode.namespaceURI != XB._base.consts.STR_FUNCTIONAL_NS) {
                        this._parseElement(treeWalker, null, parentWidget);
                        continue;
                    }
                    
                    try {
                        if (parentCalcNode == null) {
                            this._createFunctionRef(treeWalker, parentWidget);
                        }
                        else {
                            var argInfo = this._handleFunctionSubNode(treeWalker, parentWidget);
                            calcNodeArgs.push(argInfo);
                        }
                    }
                    catch (e) {
                        if (e instanceof XB.WidgetLibrary.EParseError)
                            this._logger.error(this._consts.ERR_PARSING_FUNC_NODE + " \"" + e.nodeName + "\". Reason: " + e.message);
                        else
                            throw e;
                    }
                }
                while (treeWalker.nextSibling());
            }
            finally {
                treeWalker.parentNode();
            }
        
        if (parentCalcNode != null) {
            var unnamedArgs = [];
            for each (let argInfo in calcNodeArgs) {
                if (argInfo.name)
                    parentCalcNode.attachArgument(argInfo.name, argInfo.calcNode);
                else
                    unnamedArgs.push(argInfo);
            }
            for each (let argInfo in unnamedArgs) {
                argInfo.name = parentCalcNode.proposeArgName();
                parentCalcNode.attachArgument(argInfo.name, argInfo.calcNode);
            }
        }
    },
    
    _createFunctionRef: function XBWidgetLib_createFuncRef(treeWalker, parentWidget) {
        var srcXMLNode = treeWalker.currentNode;
        
        var calcNode;
        if (srcXMLNode.localName == "value"){
            calcNode = this._createValNode(srcXMLNode);
        }
        else {
            var calcNodeConstructor = XB.functions["CN_" + srcXMLNode.localName];
            if (typeof calcNodeConstructor != "function")
                throw new XB.WidgetLibrary.EParseError(XB._base.consts.ERR_NO_RUNTIME_FUNCTION, srcXMLNode.nodeName);
            calcNode = new XB._base.FuncNodeProto(this._genNodeUID(), calcNodeConstructor);
        }
        var refID = calcNode.getBaseID();
        parentWidget.registerReference(refID, calcNode);
        
        this._logger.trace("Replacing functional element " + srcXMLNode.localName + " with a reference " + refID);
        var valRefElement = srcXMLNode.ownerDocument.createElementNS(
            XB._base.consts.STR_UI_NS,
            XB._base.consts.STR_VAL_REF_ELEM_NAME);
        valRefElement.setUserData(XB._base.consts.STR_VAL_REF_ID_KEY_NAME, refID, this._userDataHandler);
        srcXMLNode.parentNode.replaceChild(valRefElement, srcXMLNode);
        try {
            if ( !(calcNode instanceof XB._base.ConstNodeProto) ) {
                this._parseFuncNodeAttributes(srcXMLNode.attributes, calcNode);
                this._parseElement(treeWalker, calcNode, parentWidget);
            }
        }
        finally {
            treeWalker.currentNode = valRefElement;
        }
    },
    
    _handleFunctionSubNode: function XBWidgetLib_handleFuncSubNode(treeWalker, parentWidget) {
        this._logger.trace("Handling function arguments");
        
        var calcNode;
        var argName;
        var argXMLNode = treeWalker.currentNode;
        
        if (argXMLNode.nodeType == argXMLNode.ELEMENT_NODE &&
            argXMLNode.localName == this._consts.STR_FUNC_PARAM_ELEM_NAME &&
            argXMLNode.namespaceURI == XB._base.consts.STR_FUNCTIONAL_NS) {
            
            argName = argXMLNode.getAttribute("name") || "";
            this._logger.trace("Found <f:param> element, name is " + argName);
            
            var argsLen = 0;
            if (treeWalker.firstChild())
                try {
                    do {
                        argsLen++;
                    }
                    while (treeWalker.nextSibling());
                    
                    if (argsLen == 1)
                        calcNode = this._handleArgumentNode(treeWalker, parentWidget);
                }
                finally {
                    treeWalker.parentNode();
                }
            
            if (argsLen > 1) {
                var nodeID = this._genNodeUID();
                calcNode = new XB._base.FuncNodeProto(nodeID, XB.functions.CN_concat);
                this._parseElement(treeWalker, calcNode, parentWidget);
            }
            else
                if (argsLen == 0)
                    calcNode = new XB._base.ConstNodeProto(this._genNodeUID(), XB.types.empty);
        }
        else
            calcNode = this._handleArgumentNode(treeWalker, parentWidget);
        
        return {name: argName, calcNode: calcNode};
    },
    
    _handleArgumentNode: function XBWidgetLib_handleArgNode(treeWalker, parentWidget) {
        var calcNode;
        
        var argXMLNode = treeWalker.currentNode;
        if (argXMLNode.nodeType == argXMLNode.ELEMENT_NODE &&
            argXMLNode.namespaceURI == XB._base.consts.STR_FUNCTIONAL_NS) {
            this._logger.trace("Found functional parameter. Subfunction is \"" + argXMLNode.localName + "\"");
            
            if (argXMLNode.localName == "value"){
                calcNode = this._createValNode(argXMLNode);
            }
            else {
                var calcNodeConstructor = XB.functions["CN_" + argXMLNode.localName];
                if (typeof calcNodeConstructor != "function")
                    throw new XB.WidgetLibrary.EParseError(XB._base.consts.ERR_NO_RUNTIME_FUNCTION, argXMLNode.nodeName);
                calcNode = new XB._base.FuncNodeProto(this._genNodeUID(), calcNodeConstructor);
                
                this._parseFuncNodeAttributes(argXMLNode.attributes, calcNode);
                this._parseElement(treeWalker, calcNode, parentWidget);
            }
        }
        else {
            this._logger.trace("Found XML parameter. Node name is \"" + argXMLNode.nodeName + "\"");
            if (argXMLNode.nodeType == argXMLNode.TEXT_NODE) {
                this._logger.trace("Text value is: " + argXMLNode.nodeValue);
                var text = argXMLNode.nodeValue;
                var numeric = Number(text);
                calcNode = new XB._base.ConstNodeProto(this._genNodeUID(), isNaN(numeric)? text: numeric);
            }
            else {
                this._parseElement(treeWalker, null, parentWidget);
                calcNode = new XB._base.ConstNodeProto(this._genNodeUID(), new XB.types.XML(argXMLNode));
            }
        }
        
        return calcNode;
    },
    
    _parseFuncNodeAttributes: function XBWidgetLib_parseFuncNodeAttrs(attrsNodeMap, calcNode) {
        this._logger.trace("Checking functional node attributes");
        for (let attrIdx = 0, len = attrsNodeMap.length; attrIdx < len; attrIdx++) {
            var attrNode = attrsNodeMap.item(attrIdx);
            var argName = attrNode.localName;
            var arg;
            
            var text = attrNode.nodeValue;
            var match = text.match(/^\${(.+)}$/);
            if (match) {
                refName = match[1];
                refNameNode = new XB._base.ConstNodeProto(this._genNodeUID(), refName);
                arg = new XB._base.FuncNodeProto(this._genNodeUID(), XB.functions["CN_value-of"]);
                arg.attachArgument("name", refNameNode);
            }
            else {
                var trimmedArgStr = sysutils.trimString(text);
                var numeric = (trimmedArgStr.length > 0)? Number(trimmedArgStr) : NaN;
                arg = new XB._base.ConstNodeProto(this._genNodeUID(), isNaN(numeric)? text: numeric);
            }
            
            calcNode.attachArgument(argName, arg);
        }
    },
    
    _createValNode: function XBWidgetLib_createValNode(srcXMLNode) {
        var val;
        var valueType = srcXMLNode.getAttribute("type") || "string";
        switch (valueType) {
            case "string":
                val = srcXMLNode.textContent;
                break;
            case "xml":
                val = new XB.types.XML(srcXMLNode.childNodes);
                break;
            case "number":
                val = XB._base.runtime.xToNumber(srcXMLNode.textContent);
                break;
            case "bool":
                val = XB._base.runtime.xToBool(srcXMLNode.textContent);
                break;
            case "empty":
                val = XB.types.empty;
                break;
            default:
                throw new Error(this._consts.ERR_UNKNOWN_VALUE_TYPE + ": " + valueType);
        }
        return new XB._base.ConstNodeProto(this._genNodeUID(), val);
    },
    
    _genNodeUID: function XBWidgetLib_genNodeUID() {
        return this._nodeUID++;
    },
    
    _nodeFilter: {
        acceptNode : function XBWidgetLib_NF_acceptNode(node) {
            if ( (node.nodeType == node.COMMENT_NODE) ||
                 ((node.nodeType == node.TEXT_NODE) && !this._emptyTextRE.test(node.nodeValue)) )
                return XB._Ci.nsIDOMNodeFilter.FILTER_REJECT;
            return XB._Ci.nsIDOMNodeFilter.FILTER_ACCEPT;
        },
        
        _emptyTextRE: /[^\t\n\r ]/
    },
    
    _userDataHandler: {
        handle: function XBWidgetLib_UDH_handle(operation, key, data, srcNode, dstNode) {
            try {
                dstNode.setUserData(key, data, this._userDataHandler);
            }
            catch (e) {
                XB._base.logger.error("Failed setting userData " + XB._base.formatException(e));
            }
        }
    }
};

XB.WidgetLibrary.EParseError = function XBParseError(msg, nodeName) {
    this.name = "Parse error";
    this.message = msg;
    this.nodeName = nodeName;
};

XB.WidgetLibrary.EParseError.prototype = new Error();

XB.WidgetLibrary.SettingNode = XB._base.FuncNode.extend({
    constructor: function FSettingNode_constructor(baseUID, widgetInstance, prefName, defaultValue) {
        this.base(baseUID, widgetInstance);
        
        if (!prefName)
            throw new Error(this._consts.ERR_PREFNAME_EXPECTED);
        this._prefName = prefName;
        this._widgetProtoID = (widgetInstance instanceof XB.WidgetLibrary.WidgetInstance)?
            widgetInstance.getPrototype().getID():
            widgetInstance.getID();
        this._defaultValue = defaultValue || XB.types.empty;
        this._prefsModule = XB._base.application.core.Preferences;
    },
    
    finalize: function FSettingNode_finalize() {
        this._dontWrite = true;
        this.base();
    },
    
    getDefaultValue: function FSettingNode_getDefaultValue() {
        return this._defaultValue;
    },
    
    erase: function FSettingNode_erase() {
        this._erasing = true;
        try {
            this._prefsModule.reset(this._getPrefFullPath());
        }
        finally {
            this._erasing = false;
        }
    },
    
    observe: function FSettingNode_observe(subject, topic, data) {
        if (this._erasing) return;
        var prefPath = data;
        XB._base.logger.debug("Setting node " + prefPath + " observes " + topic);
        var value = this._prefsModule.get(prefPath, this._defaultValue);
        this._dontWrite = true;
        try {
            if (this._setNewVal(value))
                this._notifyDeps();
        }
        finally {
            this._dontWrite = false;
        }
    },
    
    _consts: {
        ERR_PREFNAME_EXPECTED: "Preference name expected"
    },
    _prefsModule: null,
    _prefName: undefined,
    _dontWrite: false,
    _erasing: false,
    
    _setNewVal: function FSettingNode_set(value) {
        var reset = (typeof this._storedValue) !== (typeof value);
        var changed = this.base(value);
        if (changed && !this._dontWrite) {
            var prefPath = this._getPrefFullPath();
            XB._base.logger.debug("Setting node " + prefPath + " writes new value " + value);
            this._prefsModule.ignore(prefPath, this);
            try {
                if (reset)
                    this._prefsModule.reset(prefPath);
                this._prefsModule.set(prefPath, value);
            }
            finally {
                this._prefsModule.observe(prefPath, this);
            }
        }
        return changed;
    },
    
    _notNeeded: function FSettingNode_notNeeded() {
        this._prefsModule.ignore(this._getPrefFullPath(), this);
    },
    
    _calculate: function FSettingNode_calculate() {
        var prefPath = this._getPrefFullPath();
        var value;
        if (!this._prefsModule.has(prefPath)) {
            this._prefsModule.ignore(prefPath, this);
            this._prefsModule.set(prefPath, this._defaultValue);
            value = this._defaultValue;
        }
        else
            value = this._prefsModule.get(prefPath, this._defaultValue);
        this._prefsModule.observe(prefPath, this);
        return value;
    },
    
    _getPrefFullPath: function FSettingNode_getPrefFullPath() {
        var prefPath = XB._base.application.name + ".xbwidgets." + this._widgetProtoID + ".";
        if (this._parentWidget instanceof XB.WidgetLibrary.WidgetInstance) {
            prefPath += (this._parentWidget.getID() + ".");
        }
        else
            prefPath += "all.";
        prefPath += this._prefName;
        return prefPath;
    }
});

XB.WidgetLibrary.SettingNode.implement(XB._base.IVariable);

XB.WidgetLibrary.WidgetBase = Base.extend({
    constructor: function XBWidgetBase_constructor(id) {
        if (!id)
            throw new Error(this._consts.ERR_ID_REQUIRED);
        this._id = id;
        this._varsMap = {};
        this._dataMap = {};
        this._settingsMap = {};
        this._refsMap = {};
    },
    
    getID: function XBWidgetBase_getID() {
        return this._id;
    },
    
    findReference: function XBWidgetBase_findReference(refID) {
        var refNode = this._refsMap[refID];
        return refNode? refNode: null;
    },
    
    findVariable: function XBWidgetBase_findVariable(varName) {
        var varNode = this._varsMap[varName];
        return varNode? varNode: null;
    },
    
    findData: function XBWidgetBase_findData(dataName) {
        var dataNode = this._dataMap[dataName];
        return dataNode? dataNode: null;
    },
    
    findSetting: function XBWidgetBase_findSetting(settingName) {
        var settingNode = this._settingsMap[settingName];
        return settingNode? settingNode: null;
    },
    
    eraseSettings: function XBWidgetBase_eraseSettings() {
        for each (let setting in this._settingsMap) {
            setting.erase();
        }
    },
    
    _consts : {
        WARN_VAR_REDEFINED: "Redefined variable",
        WARN_DATA_REDEFINED: "Redefined data",
        ERR_ID_REQUIRED: "ID required",
        ERR_INVALID_SCOPE: "Widget or instance scope expected"
    },
    _varsMap: null,
    _dataMap: null,
    _settingsMap: null,
    _refsMap: null
});

XB.WidgetLibrary.WidgetPrototype = XB.WidgetLibrary.WidgetBase.extend({
    constructor: function XBWidgetPrototype_constructor(srcDOMElement, library) {
        if ( !(srcDOMElement instanceof XB._Ci.nsIDOMElement) )
            throw new TypeError(XB._base.consts.ERR_DOM_NODES_EXPECTED);
        this._srcDOMElement = srcDOMElement;
        var protoID = srcDOMElement.getAttribute("id");
        this.base(protoID);
        if ( !(library instanceof XB.WidgetLibrary) )
            throw new TypeError(XB._base.consts.ERR_WLIB_REQUIRED);
        this._library = library;
        this._unique = (srcDOMElement.getAttribute("unique") == "true");
        this._name = srcDOMElement.getAttribute("name") || undefined;
        this._iconPath = srcDOMElement.getAttribute("icon") || undefined;
        this._instSettings = [];
        this._instVars = [];
    },

    createInstance: function XBWidgetPrototype_createInstance(instanceID, windowEngine, instanceSettings) {
        var instance = new XB.WidgetLibrary.WidgetInstance(instanceID,
                                                           this,
                                                           windowEngine,
                                                           this._refsMap,
                                                           this._dataMap,
                                                           this._instSettings,
                                                           this._instVars);
        
        if (instanceSettings != undefined)
            XB._base.logger.debug("Instance settings:\n" + misc.dump(instanceSettings));
            try {
                for (let name in instanceSettings)
                    instance.applySetting(name, instanceSettings[name]);
            }
            catch (e) {
                XB._base.logger.error("Couldn't apply widget instance settings. " + XB._base.formatException(e));
            }
        
        return instance;
    },
    
    getSrcElement: function XBWidgetPrototype_getSrcElement() {
        return this._srcDOMElement;
    },
    
    registerReference: function XBWIdgetPrototype_registerReference(refID, calcNode) {
        this._refsMap[refID] = calcNode;
    },
    
    registerVariable: function XBWidgetPrototype_registerVariable(varName, varScope, persist, initialValue) {
        XB._base.logger.debug("Widget prototype " + this.getID() + " registers var " + varName + ", scope " + varScope);
        if (this._varsMap[varName])
            XB._base.logger.warn(this._consts.WARN_VAR_REDEFINED + " " + varName);
        
        switch (varScope) {
            case XB._base.consts.ENUM_SCOPE_WIDGET:
                var nodeID = this.getID() + "_var_" + varName;
                this._varsMap[varName] = new XB._base.VarNode(nodeID, this, initialValue);
                break;
            case XB._base.consts.ENUM_SCOPE_INSTANCE:
                this._instVars.push({name: varName, initialValue: initialValue});
                break;
            default:
                throw new Error(this._consts.ERR_INVALID_SCOPE);
        }
    },
    
    registerData: function XBWidgetPrototype_registerData(dataName, dataNode) {
        if (this._dataMap[dataName])
            XB._base.logger.warn(this._consts.WARN_DATA_REDEFINED + " " + dataName);
        this._dataMap[dataName] = dataNode;
    },
    
    registerSetting: function XBWidgetPrototype_registerSetting(settingName, settingScope, defaultValue) {
        XB._base.logger.debug("Widget prototype " + this.getID() + " registers setting " + settingName + ", scope " + settingScope);
        
        switch (settingScope) {
            case XB._base.consts.ENUM_SCOPE_WIDGET:
                var nodeID = "setting_" + settingName;
                var settingNode = new XB.WidgetLibrary.SettingNode(nodeID, this, settingName, defaultValue);
                this._settingsMap[settingName] = settingNode;
                break;
            case XB._base.consts.ENUM_SCOPE_INSTANCE:
                this._instSettings.push({name: settingName, defaultValue: defaultValue});
                break;
            default:
                throw new Error(this._consts.ERR_INVALID_SCOPE);
        }
    },
    
    getName: function XBWidgetPrototype_getName() {
        return this._name;
    },
    
    getIconPath: function XBWidgetPrototype_getIconPath() {
        return this._iconPath;
    },

    isUnique: function XBWidgetPrototype_isUnique() {
        return this._unique;
    },
    
    getLibrary: function XBWidgetPrototype_getLibrary() {
        return this._library;
    },
    
    _srcDOMElement: undefined,
    _name: undefined,
    _iconPath: undefined,
    _library: null,
    _unique: false,
    _instSettings: null,
    _instVars: null
});

XB.WidgetLibrary.WidgetInstance = XB.WidgetLibrary.WidgetBase.extend({
    constructor: function XBWidgetInstance_constructor(IID, proto, wndEngine, refsMap, dataMap, settingsArr, varsArr) {
        if ( !(proto instanceof XB.WidgetLibrary.WidgetPrototype) )
            throw new TypeError(this._consts.ERR_PROTO_EXPECTED);
        this.base(IID);
        this._proto = proto;
        this._wndEngine = wndEngine;
        
        for (let refID in refsMap)
            this._refsMap[refID] = refsMap[refID].createInstance(this);
        
        for (let dataName in dataMap)
            this._dataMap[dataName] = dataMap[dataName].createInstance(this);
        
        for each(let setting in settingsArr)
            this._registerSetting(setting.name, setting.defaultValue);
            
        for each(let variable in varsArr)
            this._registerVariable(variable.name, variable.initialValue);
    },
    
    getPrototype: function XBWidgetInstance_getPrototype() {
        return this._proto;
    },
    
    getWindowEngine: function XBWidgetInstance_getWindowEngine() {
        return this._wndEngine;
    },
    
    applySetting: function XBWidgetInstance_applySetting(name, value) {
        this._settingsMap[name].setValue(value);
    },
    
    findSetting: function XBWidgetInstance_findSetting(settingName) {
        this._wndEngine.logger.debug("Looking for widget instance setting " + settingName);
        var settingNode = this.base(settingName);
        return settingNode? settingNode: this._proto.findSetting(settingName);
    },
    
    findVariable: function XBWidgetInstance_findVariable(varName) {
        this._wndEngine.logger.debug("Looking for widget instance variable " + varName);
        var varNode = this.base(varName);
        return varNode? varNode: this._proto.findVariable(varName);
    },
    
    finalize: function XBWidgetInstance_finalize() {
        function finalizeNodesInMap(map) {
            for each (let node in map)
                try {
                    node.finalize();
                }
                catch (e) {}
        }
        
        finalizeNodesInMap(this._refsMap);
        finalizeNodesInMap(this._varsMap);
        finalizeNodesInMap(this._dataMap);
        finalizeNodesInMap(this._settingsMap);
        
        this._refsMap = null;
        this._varsMap = null;
        this._dataMap = null;
        this._settingsMap = null;
        
        //this._wndEngine = null;
        //this._proto = null;
    },
    
    _consts: {
        ERR_PROTO_EXPECTED: "Widget prototype expected"
    },
    _proto: null,
    _wndEngine: null,
    
    _registerSetting: function XBWidgetInstance_registerSetting(settingName, defaultValue) {
        var nodeID = "setting_" + settingName;
        var settingNode = new XB.WidgetLibrary.SettingNode(nodeID, this, settingName, defaultValue);
        this._settingsMap[settingName] = settingNode;
    },
    
    _registerVariable: function XBWidgetInstance_registerVariable(varName, initialValue) {
        var nodeID = this.getID() + "_var_" + varName;
        this._varsMap[varName] = new XB._base.VarNode(nodeID, this, initialValue);
    }
});
