XB.WindowEngine = function XBWndEngine_constructor(widgetsRegistry, overlayController) {
    if ( !(widgetsRegistry instanceof XB.WidgetsRegistry) )
        throw new TypeError(this._consts.ERR_WREG_EXPECTED);
    
    this._eid = "WndEngine" + XB._base.genWindowEngineID();
    this._logger = XB._base.application.core.Log4Moz.repository.getLogger(XB._base.loggersRoot + "." + this._eid);
    this._widgetsMap = {};
    this._overlayController = overlayController;
    
    this._widgetsRegistry = widgetsRegistry;
};


XB.WindowEngine.prototype = {
    constructor: XB.WindowEngine,
    
    get window() {
        return this._overlayController.window;
    },
    
    clear: function XBWndEngine_clear() {
        for each (let widget in this._widgetsMap) {
            try {
                widget.finalize();
            }
            catch (e) {}
        }
        this._widgetsMap = {};
    },
    
    hasWidget: function XBWndEngine_hasWidget(widgetIID) {
        return !!(this._widgetsMap[widgetIID]);
    },
    
    createWidget: function XBWndEngine_createWidget(protoID, instID, settings) {
        if (!instID)
            throw new Error(this._consts.ERR_INSTID_REQUIRED);
        if (this.hasWidget(instID)) {
            this._logger.warn("Widget with this ID (" + instID + ") already exists. Will return old one.");
            let widget = this.getWidget(instID);
            if (settings) {
                for (let settingName in settings) {
                    widget.applySetting(settingName, settings[settingName]);
                }
            }
            return widget;
        }
        
        var widgetProto = this._widgetsRegistry.findProto(protoID);
        var widget = widgetProto.createInstance(instID, this, settings);
        this._widgetsMap[instID] = widget;
        return widget;
    },
    
    getWidget: function XBWndEngine_getWidget(widgetIID) {
        var widget = this._widgetsMap[widgetIID];
        if (!widget)
            throw new Error(this._consts.ERR_WIDGET_NOT_FOUND + " (" + widgetIID + ")");
        return widget;
    },
    
    getValueAsString: function XBWndEngine_getValueAsString(widgetIID, valUID) {
        return this._safeGetStrNodeValue(this._getCalcNode(widgetIID, valUID));
    },
    
    getValue: function XBWndEngine_getValue(widgetIID, valUID) {
        return this._getCalcNode(widgetIID, valUID).getValue(this);
    },
    
    perform: function XBWndEngine_perform(widgetIID, valUID) {
        try {
            var procNode = this._getCalcNode(widgetIID, valUID);
        }
        catch (e) {
            throw new XB.WindowEngine.ENoActionHandler(e.message);
        }
        if ( !(procNode instanceof XB._base.ProcNode) )
            throw new TypeError(XB._base.consts.ERR_PROC_NODE_EXPECTED);
        
        procNode.perform();
    },
    
    valueNotNeeded: function XBWndEngine_valueNotNeeded(widgetIID, valUID) {
        this._getCalcNode(widgetIID, valUID).unsubscribe(this);
    },
    
    freeze: function XBWndEngine_freeze() {
        
    },
    
    melt: function XBWndEngine_melt(changedNode) {
        if (!changedNode) return;
        try {
            this._notifyUIBuilder(changedNode);
        }
        catch (e) {
            this._logger.error("Failed notifying UI builder. " + XB._base.formatException(e));
            if (e.stack)
                this._logger.debug(e.stack);
        }
    },
    
    getEffectiveID: function XBWndEngine_getEffectiveID() {
        return this.id;
    },
    
    get id() {
        return this._eid;
    },
    
    get logger() {
        return this._logger;
    },
    
    _consts: {
        ERR_INSTID_REQUIRED: "Instance ID required",
        ERR_XF_NOT_FOUND: "Requested value not found",
        ERR_WIDGET_NOT_FOUND: "Couldn't find widget with this ID",
        ERR_WI_EXPECTED: "Widgets info array expected",
        ERR_LIBRARY_EXPECTED: "Widget library instance expected",
        ERR_WREG_EXPECTED: "Widget registry expected",
        ERR_CANT_CONSTRUCT_WIDGET: "Failed creating widget"
    },
    _logger: XB._noopLogger,
    _eid: undefined,
    _overlayController: null,
    _widgetsMap: null,
    _widgetsRegistry: null,
    
    _getCalcNode: function XBWndEngine__getCalcNode(widgetIID, valUID) {
        var calcNode =  this.getWidget(widgetIID).findReference(valUID);
        if (!calcNode)
            throw new Error(this._consts.ERR_XF_NOT_FOUND + " (" + [widgetIID, valUID] + ")");
        return calcNode;
    },
    
    _safeGetStrNodeValue: function XBWndEngine__safeGetStrNodeValue(node) {
        var strVal = "";
        try {
            strVal = XB._base.runtime.xToString(node.getValue(this));
        }
        catch (e) {
            this._logger.error("Converting value to string failed. " + XB._base.formatException(e) + ". Will return empty string.");
        }
        return strVal;
    },
    
    _notifyUIBuilder: function XBWndEngine__notifyUIBuilder(changedNode) {
        this._logger.debug("Notifying UI builder");
        
        var baseID = changedNode.getBaseID();
        var widgetIID = changedNode.getParentWidget().getID();
        this._overlayController.guiBuilder.handleDataChange(widgetIID, baseID, changedNode.getValue(this));
    }
};

XB.WindowEngine.ENoActionHandler = function XBWE_ENoActionHandler(msg) {
    this.name = "No action handler error";
    this.message = msg;
};

XB.WindowEngine.ENoActionHandler.prototype = new Error();