const APP_NAME = "yasearch";
const EXT_ID = "yasearch@yandex.ru";
const CORE_GUID_STR = "{F25B83DE-5817-11DE-8EB3-C9A656D89593}";
const CORE_CONTRACTID = "@yandex.ru/custombarcore;" + APP_NAME;
const PROTOCOL_CLASSID_STR = "{1A2ADED0-64D1-11DE-8047-DBAA3A7E4459}";

const PROTOCOL_XB_CLASSID = Components.ID("{6BA7D0A8-A115-11DE-8D96-029555D89593}");
const PROTOCOL_XB_SCHEME = "xb";
const PROTOCOL_XB_CONTRACTID = "@mozilla.org/network/protocol;1?name=" + PROTOCOL_XB_SCHEME;

const Cc = Components.classes,
    Ci = Components.interfaces,
    Cr = Components.results,
    Cu = Components.utils;
const CommonConsts = {
    APP_ACTION_QUIT_TOPIC: "quit-application",
    PROFILE_AFTER_CHANGE: "profile-after-change",
    PROFILE_BEFORE_CHANGE: "profile-before-change"
};

Cu.import("resource://gre/modules/XPCOMUtils.jsm");

function CustomBarCore() {
    Cu.import(this._modulesPath + "log4moz.jsm", this);
    Cu.import(this._modulesPath + "Preferences.jsm", this);
    Cu.import(this._modulesPath + "JSON.jsm", this);
    
    this._setupLogging();
    this._logger.info("Hello from " + this.contractID);
    
    this._importDependencies(this._libFiles);
    this._observerService.addObserver(this, CommonConsts.PROFILE_AFTER_CHANGE, false);
    this._observerService.addObserver(this, CommonConsts.PROFILE_BEFORE_CHANGE, false);
    this._observerService.addObserver(this, CommonConsts.APP_ACTION_QUIT_TOPIC, false);
};

CustomBarCore.prototype = {
    Lib: {},
    Log4Moz: null,
    Preferences: null,
    JSON: null,
    
    formatExceptionMessage: function CustomBarCore_formatException(e) {
        var text = e.name + ": " + e.message;
        if (e.fileName)
            text += "\nin " + e.fileName + "\nat line " + e.lineNumber;
        return  text;
    },
    
    get appName() {
        return this._appName;
    },
    
    get application() {
        return this._appObj;
    },
    
    get extensionPathFile() {
        return Cc["@mozilla.org/extensions/manager;1"]
            .getService(Ci.nsIExtensionManager)
                .getInstallLocation(EXT_ID)
                    .getItemLocation(EXT_ID);
    },
    
    get wrappedJSObject() {
        return this;
    },
    
    observe: function CustomBarCore_observe(subject, topic, data) {
        this._logger.debug("Observing on topic \"" + topic + "\" with subject \"" + subject + "\"");
        this._logger.debug("Data is:\n" + data);
        
        switch(topic) {
            case CommonConsts.PROFILE_AFTER_CHANGE:
                try {
                    this._updateLoggers(true, true, true);
                    this.Preferences.observe(this._dumpLogLevelPrefName, this);
                    this.Preferences.observe(this._consoleLogLevelPrefName, this);
                }
                finally {
                    this._initApp();
                }
                break;
            case CommonConsts.PROFILE_BEFORE_CHANGE:
                this._destroyApp();
                break;
            case CommonConsts.APP_ACTION_QUIT_TOPIC:
                this._shutdown();
                break;
            case "nsPref:changed":
                this._updateLoggers(data == this._dumpLogLevelPrefName, data == this._consoleLogLevelPrefName);
                break;
        }
    },
    
    _libFiles: ["Base.js", "sysutils.js",
                "overlay.js", "misc.js",
                "xbbase.js", "xbtypes.js", "xbcalcnodes.js", "xbwlib.js", "xbpackage.js", "xbdb.js", "xbcache.js",
                "xbnetwork.js", "xbfuncs.js", "xbpreset.js", "xbwndengine.js", "xulbuilder.js"],
    _modulesPath: "resource://" + APP_NAME + "/",
    _appChromePath: "chrome://" + APP_NAME + "/",
    _dumpLogLevelPrefName: APP_NAME + ".xbcore.logging.stdout.level",
    _consoleLogLevelPrefName: APP_NAME + ".xbcore.logging.console.level",
    _appName: APP_NAME,
    _appObj: null,
    _dumpAppender: null,
    _consoleAppender: null,
    _observerService: Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService),
    
    _setupLogging: function CustomBarCore_setupLogging() {
        let root = this.Log4Moz.repository.rootLogger;
        root.level = this.Log4Moz.Level.All;
        
        let formatter = new this.Log4Moz.BasicFormatter();
        
        this._consoleAppender = new this.Log4Moz.ConsoleAppender(formatter);
        this._consoleAppender.level = 100;
        root.addAppender(this._consoleAppender);
        
        this._dumpAppender = new this.Log4Moz.DumpAppender(formatter);
        this._dumpAppender.level = 100;
        root.addAppender(this._dumpAppender);
        
        this._logger = this.Log4Moz.repository.getLogger(this.appName + ".Core");
        this._logger.level = this.Log4Moz.Level.Debug;
    },
    
    _updateLoggers: function CustomBarCore_updateLoggers(checkDumpSetting, checkConsoleSetting) {
        if (checkDumpSetting) {
            var dumpLevel = parseInt( this.Preferences.get(this._dumpLogLevelPrefName, -1) );
            if (!isNaN(dumpLevel)) {
                if (dumpLevel < 0)
                    dumpLevel = 100;
                this._dumpAppender.level = dumpLevel;
            }
        }
        
        if (checkConsoleSetting) {
            var consoleLevel = parseInt( this.Preferences.get(this._consoleLogLevelPrefName, -1) );
            if (!isNaN(consoleLevel)) {
                if (consoleLevel < 0)
                    consoleLevel = 100;
                this._consoleAppender.level = consoleLevel;
            }
        }
    },
    
    _importDependencies: function CustomBarCore_importDependencies(modulesNames) {
        if ((typeof modulesNames !== "object") || (modulesNames.constructor !== Array))
            throw new TypeError("Array expected as a second parameter");
        
        var libChromePath =  this._appChromePath + "content/custombar/lib/";
        var mozSSLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
        
        for each (let moduleName in modulesNames) {
            this._logger.config("Loading module from " + moduleName);
            try {
                mozSSLoader.loadSubScript(libChromePath + moduleName, this.Lib);
            }
            catch (e) {
                this._logger.fatal(
                    "Failed loading module from " + moduleName + "\n" + this.formatExceptionMessage(e));
                this._logger.debug("Stack trace:\n" + e.stack);
            }
        };
    },
    
    _initApp: function CustomBarCore_initApp() {
        var appScript = {};
        try {
            var mozSSLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
            mozSSLoader.loadSubScript(this._appChromePath + "content/custombar/bar.js", appScript);
            this._appObj = new (appScript[this.appName])(this);
        }
        catch (e) {
            this._logger.fatal("Couldn't initialize application. " + this.formatExceptionMessage(e));
        }
    },
    
    _destroyApp: function CustomBarCore_destroyApp() {
        try {
            this._appObj.finalize();
            this._appObj = null;
        } catch(e) {
            this._logger.error(this.formatExceptionMessage(e));
        }
    },
    
    _shutdown: function CustomBarCore_shutdown(){
    },
    
    classDescription: "Custom Yandex bar core JS component for " + APP_NAME,
    classID: Components.ID(CORE_GUID_STR),
    contractID: "@yandex.ru/custombarcore;" + APP_NAME,
    QueryInterface: XPCOMUtils.generateQI(),
    _xpcom_factory: {
        createInstance: function (outer, iid) {
            if (outer != null)
                throw Components.results.NS_ERROR_NO_AGGREGATION;
            if (!this._componentInstance)
                this._componentInstance = new CustomBarCore();
            return this._componentInstance.QueryInterface(iid);
        }
    },
    _xpcom_categories: [
        {
            category: "app-startup",
            service: true
        }
    ]
};

function CustomBarProtocol() {
    this._providers = [];
    this._observerService.addObserver(this, CommonConsts.APP_ACTION_QUIT_TOPIC, false);
};

CustomBarProtocol.prototype = {
    addDataProvider: function CBPH_addDataProvider(aProvider) {
        if (this._providers.some(function(handler) { return aProvider === handler; }))
            return;
        
        this._providers.push(aProvider);
    },
    
    removeDataProvider: function CBPH_removeDataProvider(aProvider) {
        this._providers = this._providers.filter(function(handler) { return aProvider !== handler; });
    },
    
    get wrappedJSObject() {
        return this;
    },
    
    observe: function CBPH_observe(aSubject, aTopic, aData) {
        switch (aTopic) {
            case CommonConsts.APP_ACTION_QUIT_TOPIC:
                this._destroy();
                break;
        }
    },
    
    get scheme() {
        return this._scheme;
    },
    
    get protocolFlags() {
        return Ci.nsIProtocolHandler.URI_DANGEROUS_TO_LOAD;
    },
    
    get defaultPort() {
        return -1;
    },
    
    allowPort: function CBPH_allowPort(port, scheme) {
        return false;
    },
    
    newURI: function CBPH_newURI(aSpec, aOriginalCharset, baseURI) {
        let uri = this._handleNewURI(aSpec, aOriginalCharset, baseURI);
        
        if (!uri)
            throw Cr.NS_ERROR_MALFORMED_URI;
        
        return uri;
    },
    
    newChannel: function CBPH_newChannel(aURI) {
        var channel = this._ioService.newChannel("data:,", null, null);
        channel.originalURI = aURI;
        return channel;
    },
    
    _scheme: APP_NAME,
    _ioService: Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService),
    _observerService: Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService),
    
    _handleNewURI: function CBPH__handleNewURI(aSpec, aOriginalCharset, aBaseURI) {
        let uri = null;
        let spec = aSpec.replace(this.scheme + ":", "");
        
        this._providers.some(function(aProvider) {
            return (null !== (uri = aProvider.newURI(spec, aOriginalCharset, aBaseURI)));
        });
        
        if (uri && typeof uri == "string")
            uri = this._ioService.newURI(uri, aOriginalCharset, aBaseURI);
        
        return uri;
    },
    
    _destroy: function CBPH__destroy() {
        this._observerService.removeObserver(this, CommonConsts.APP_ACTION_QUIT_TOPIC, false);
        this._providers.forEach(function(aProvider) aProvider.clear());
        this._providers = null;
    },
    
    classDescription: "Custom Yandex bar protocol handler JS component for " + APP_NAME,
    classID: Components.ID(PROTOCOL_CLASSID_STR),
    contractID: "@mozilla.org/network/protocol;1?name=" + APP_NAME,
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIProtocolHandler]),
    
    _xpcom_categories: [{
        category: "app-startup",
        service: true
    }]
};

function XBURI(aSpec, aOriginalCharset, aBaseURI) {
    if (aBaseURI)
        this.spec = aBaseURI.resolve(aSpec);
    else
        this.spec = aSpec;
}

XBURI.prototype = {
    get wrappedJSObject() {
        return this;
    },
    
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIURI, Ci.nsISupports]),
    
    get provider() {
        return this._provider;
    },
    
    get spec() {
        return this._spec;
    },
    
    set spec(aSpec) {
        this._spec = aSpec;
        this._parse();
        return this._spec;
    },
    
    prePath: "",
    scheme: PROTOCOL_XB_SCHEME,
    userPass: "",
    username: "",
    password: "",
    hostPort: "",
    
    get host() {
        return this._host;
    },
    
    set host(aHost) {
        this._host = aHost;
        this._construct();
        return this._host;
    },
    
    port: -1,
    
    get path() {
        return this._path;
    },
    set path(aPath) {
        this._path = aPath;
        this._construct();
        return this._path;
    },
    get asciiSpec() {
        return this.spec;
    },
    get asciiHost() {
        return this.host;
    },
    get originCharset() {
        return "UTF-8";
    },
    
    clone: function() {
        return new XBURI(this.spec);
    },
    
    equals: function(aUri) {
        return aUri.spec == this.spec;
    },
    
    resolve: function(aRelativePath) {
        if (aRelativePath[0] == "/")
            aRelativePath = this.scheme + '://' + aRelativePath;
        aRelativePath += ";" + this.provider;
        return aRelativePath;
    },
    
    schemeIs: function(aScheme) {
        aScheme = aScheme.toLowerCase();
        return aScheme == this.scheme || aScheme == "chrome";
    },
    
    _host: "",
    _path: "",
    _provider: "",
    _spec: "",
    _reSpec: /^(\w+)\:\/\/(\w*)(\/[^\?\;\#]*);(.*)$/,
    
    _parse: function() {
        var components = this._spec.match(this._reSpec);
        if (!components || components[1] != this.scheme)
            throw Cr.NS_ERROR_MALFORMED_URI;
        this._host = components[2];
        this._path = components[3];
        this._provider = components[4];
    },
    
    _construct: function() {
        var query = this.query;
        var spec = this.scheme + "://" + this.host + "/" + this.path + ";" + this.provider;
    }
}

function XBProtocolHandler() {
    this._core = Components.classes[CORE_CONTRACTID].getService().wrappedJSObject;
    this._providers = {};
    this._observerService.addObserver(this, CommonConsts.APP_ACTION_QUIT_TOPIC, false);
};

XBProtocolHandler.prototype = {
    classDescription: "Custom Yandex.Bar protocol handler for XB-resources.",
    classID: PROTOCOL_XB_CLASSID,
    contractID: PROTOCOL_XB_CONTRACTID,
    QueryInterface: XPCOMUtils.generateQI([Ci.nsIProtocolHandler, Ci.nsISupports, Ci.nsIObserver]),
    
    _xpcom_categories: [{
        category: "app-startup",
        service: false
    }],
    
    addDataProvider: function XBProtocolHandler_addDataProvider(aProvider) {
        this._providers[aProvider.providerId] = aProvider;
    },
 
    removeDataProvider: function XBProtocolHandler_removeDataProvider(aProvider) {
        delete this._providers[aProvider.providerId];
    },
 
    get wrappedJSObject() {
        return this;
    },
    
    observe: function XBProtocolHandler_observe(aSubject, aTopic, aData) {
    },
    
    get scheme() {
        return PROTOCOL_XB_SCHEME;
    },
    
    get protocolFlags() {
        return Ci.nsIProtocolHandler.URI_DANGEROUS_TO_LOAD | Ci.nsIProtocolHandler.URI_NOAUTH;
    },
    
    get defaultPort() {
        return -1;
    },
    
    allowPort: function XBProtocolHandler_allowPort() {
        return false;
    },
    
    newURI: function XBProtocolHandler_newURI(aSpec, aOriginalCharset, aBaseURI) {
        return new XBURI(aSpec, aOriginalCharset, aBaseURI);
    },
    
    newChannel: function XBProtocolHandler_newChannel(aURI) {
        var uri = new XBURI(aURI.spec);
        var channel = null;
        
        var provider = this._providers[uri.provider];
        if (provider)
            channel = provider.newChannel(uri);
        if (!channel)
            throw Components.results.NS_ERROR_FAILURE;
        return channel;
    },
    
    _observerService: Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService),
    
    _destroy: function XBProtocolHandler__destroy() {
        this._observerService.removeObserver(this, CommonConsts.APP_ACTION_QUIT_TOPIC, false);
        this._providers = null;
    }
};

var components = [CustomBarCore, CustomBarProtocol];

/* if (!(PROTOCOL_XB_CONTRACTID in Cc)) */
  components.push(XBProtocolHandler);


function NSGetModule(compMgr, fileSpec) {
    return XPCOMUtils.generateModule(components);
}